library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity top is
port(
--INPUTS
F14         : in std_logic;
nRES        : in std_logic;
PORT_FE_D5  : in std_logic;
--OUTPUTS
CLK         : out std_logic;
VA          : out std_logic_vector(15 downto 0);
ZA          : out std_logic_vector(15 downto 0);
ZD          : out std_logic_vector(7 downto 0);
nRD         : out std_logic;
nWR         : out std_logic;
nMREQ       : out std_logic;
nIORQ       : out std_logic;
nM1         : out std_logic;
BORDER      : out std_logic;
CSYNC       : out std_logic;
KSYNC       : out std_logic;
BLANK       : out std_logic;
FLASH       : out std_logic;
ZZ80_nINT   : out std_logic;
ZZ80_nRES   : out std_logic
);
end;

architecture top_rtl of top is

component cnt_D2 is
    port(
        C   : in std_logic;
        q   : out std_logic_vector(4 downto 1);
        PO  : out std_logic
    );
end component;

component cnt_D4 is
    port(
        C   : in std_logic;
        nE  : in std_logic;
        q   : out std_logic_vector(5 downto 1)
    );
end component;

component vrom is
    port(
        address : in std_logic_vector(10 downto 0);
        clock   : in std_logic;
        q       : out std_logic_vector(7 downto 0)
    );
end component;

component T80s is
	generic(
		Mode : integer := 0;	-- 0 => Z80, 1 => Fast Z80, 2 => 8080, 3 => GB
		T2Write : integer := 0;	-- 0 => WR_n active in T3, /=0 => WR_n active in T2
		IOWait : integer := 1	-- 0 => Single cycle I/O, 1 => Std I/O cycle
	);
	port(
		RESET_n		: in std_logic;
		CLK_n		: in std_logic;
		WAIT_n		: in std_logic;
		INT_n		: in std_logic;
		NMI_n		: in std_logic;
		BUSRQ_n		: in std_logic;
		M1_n		: out std_logic;
		MREQ_n		: out std_logic;
		IORQ_n		: out std_logic;
		RD_n		: out std_logic;
		WR_n		: out std_logic;
		RFSH_n		: out std_logic;
		HALT_n		: out std_logic;
		BUSAK_n		: out std_logic;
		A			: out std_logic_vector(15 downto 0);
		DI			: in std_logic_vector(7 downto 0);
		DO			: out std_logic_vector(7 downto 0);
		RestorePC   : in std_logic_vector(15 downto 0);
        RestoreINT  : in std_logic_vector(7 downto 0);
		RestorePC_n : in std_logic
	);
end component;

signal PE           : std_logic;
signal Q            : std_logic_vector(4 downto 1);
signal C            : std_logic_vector(5 downto 1);

signal VROM_ADDR    : std_logic_vector(10 downto 0);
signal VROM_DO      : std_logic_vector(7 downto 0);

signal Z80_ADDR     : std_logic_vector(15 downto 0):="0000000000000000";
signal Z80_DI       : std_logic_vector(7 downto 0);
signal Z80_DO       : std_logic_vector(7 downto 0);
signal Z80_nWR      : std_logic;
signal Z80_nRD      : std_logic;
signal Z80_nMREQ    : std_logic;
signal Z80_nIORQ    : std_logic;
signal Z80_nM1      : std_logic;
signal Z80_nINT     : std_logic;
signal Z80_nNMI     : std_logic;
signal Z80_nHALT    : std_logic;
signal Z80_nCLK     : std_logic;

signal REG_D10_WR   : std_logic;
signal REG_D11_WR   : std_logic;

signal REG_D10      : std_logic_vector(7 downto 0);
signal REG_D11      : std_logic_vector(7 downto 0);

begin
        VA(4 downto 0)  <= C;
        VA(7 downto 5)  <= REG_D10(5 downto 3);
        VA(15 downto 8) <= REG_D11(5 downto 3)&REG_D11(7 downto 6)&REG_D11(2 downto 0) when Q(4) = '1' else "010110"&REG_D10(7 downto 6);

        BORDER      <= REG_D10(1);
        CSYNC       <= REG_D10(0);
        KSYNC       <= Z80_ADDR(12);
        BLANK       <= Z80_ADDR(14);
        FLASH       <= Z80_ADDR(15);

        ZA          <= Z80_ADDR;
        ZD          <= Z80_DI;
        nRD         <= Z80_nRD;
        nWR         <= Z80_nWR;
        nMREQ       <= Z80_nMREQ;
        nIORQ       <= Z80_nIORQ;
        nM1         <= Z80_nM1;

        ZZ80_nINT   <= REG_D10(2);
        ZZ80_nRES   <= Z80_nHALT;   
        
        Z80_nINT    <= C(5);
        Z80_nNMI    <= Z80_ADDR(11);
        Z80_nCLK    <= not Q(2);
        CLK         <= not Q(2);
        VROM_ADDR   <= Z80_nM1&Z80_ADDR(9 downto 0);

        REG_D10_WR  <= Z80_nMREQ or Z80_nWR;
        REG_D11_WR  <= Z80_nIORQ or Z80_nWR;

        process(Z80_nMREQ,Z80_nRD,Z80_nIORQ)
        begin
            if Z80_nRD = '0' and Z80_nMREQ = '0' then
                Z80_DI <= VROM_DO;
            elsif Z80_nRD = '0' and Z80_nIORQ = '0' then
                Z80_DI <= "111111"&PORT_FE_D5&not Q(4);
            else
                Z80_DI <= "11111111";
            end if;
        end process;

        process(REG_D10_WR)
        begin
            if REG_D10_WR'event and REG_D10_WR = '1' then
                REG_D10 <= Z80_DO;
            end if;
        end process;
        
        process(REG_D11_WR)
        begin
            if REG_D11_WR'event and REG_D11_WR = '1' then
                REG_D11 <= Z80_DO;
            end if;
        end process;

        cntD2: cnt_D2
        PORT MAP( C => F14, q => Q, PO => PE);

        cntD4: cnt_D4
        PORT MAP( C => PE, nE =>Z80_ADDR(13), q => C);

        Z80:T80s
        PORT MAP(
            RESET_n	=>nRES,
            CLK_n	=>Z80_nCLK,
            WAIT_n	=>'1',
            INT_n	=>Z80_nINT,
            NMI_n	=>Z80_nNMI,
            BUSRQ_n	=>'1',
            M1_n	=>Z80_nM1,
            MREQ_n	=>Z80_nMREQ,
            IORQ_n	=>Z80_nIORQ,
            RD_n	=>Z80_nRD,
            WR_n	=>Z80_nWR,
            RFSH_n	=>open,
            HALT_n	=>Z80_nHALT,
            BUSAK_n	=>open,
            A		=>Z80_ADDR,
            DI		=>Z80_DI,
            DO		=>Z80_DO,
            RestorePC_n =>'1',
            RestorePC   =>"1111111111111111",
            RestoreINT  =>"11111111"
        );

        vrom1 : vrom
        PORT MAP (
            address => VROM_ADDR,
            clock	=> F14,
            q	    => VROM_DO
        );

end top_rtl;